# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
r"""
Lie Conformal Algebra

Let `R` be a commutative ring, a *super Lie conformal algebra*
[Kac1997]_ over `R`
(also known as a *vertex Lie algebra*) is an `R[T]` super module `L`
together with a `\mathbb{Z}/2\mathbb{Z}`-graded `R`-bilinear
operation (called the `\lambda`-bracket)
`L\otimes L \rightarrow L[\lambda]`
(polynomials in `\lambda` with
coefficients in `L`), `a \otimes b \mapsto [a_\lambda b]` satisfying

1. Sesquilinearity:

   .. MATH::

        [Ta_\lambda b] = - \lambda [a_\lambda b], \qquad [a_\lambda Tb] =
        (\lambda+ T) [a_\lambda b].

2. Skew-Symmetry:

   .. MATH::

        [a_\lambda b] = - (-1)^{p(a)p(b)} [b_{-\lambda - T} a],

   where `p(a)` is `0` if `a` is *even* and `1` if `a` is *odd*. The
   bracket in the RHS is computed as follows. First we evaluate
   `[b_\mu a]` with the formal
   parameter `\mu` to the *left*, then
   replace each appearance of the formal variable `\mu` by `-\lambda - T`.
   Finally apply `T` to the coefficients in `L`.

3. Jacobi identity:

   .. MATH::

       [a_\lambda [b_\mu c]] = [ [a_{\lambda + \mu} b]_\mu c] +
       (-1)^{p(a)p(b)} [b_\mu [a_\lambda c ]],

   which is understood as an equality in `L[\lambda,\mu]`.

   `T` is usually called the *translation operation* or the *derivative*.
   For an element `a \in L` we will say that `Ta` is the *derivative of*
   `a`. We define the *`n`-th products* `a_{(n)} b` for `a,b \in L` by

   .. MATH::

        [a_\lambda b] = \sum_{n \geq 0} \frac{\lambda^n}{n!} a_{(n)} b.

   A Lie conformal algebra is called *H-Graded* [DSK2006]_ if there exists
   a decomposition `L = \oplus L_n` such that the
   `\lambda`-bracket becomes graded of degree `-1`, that is:

   .. MATH::

        a_{(n)} b \in L_{p + q -n -1} \qquad
        a \in L_p, \: b \in L_q, \: n \geq 0.

   In particular this implies that the action of `T` increases
   degree by `1`.

.. NOTE::

    In the literature arbitrary gradings are allowed. In this
    implementation we only support nonnegative rational gradings.

EXAMPLES:

1. The **Virasoro** Lie conformal algebra `Vir` over a ring `R`
   where `12` is invertible has two generators `L, C` as an `R[T]`-module.
   It is the direct sum of a free module of rank `1` generated by `L`, and
   a free rank one `R` module generated by `C` satisfying `TC = 0`.  `C`
   is central (the `\lambda`-bracket of `C` with any other vector
   vanishes). The remaining `\lambda`-bracket is given by

   .. MATH::

        [L_\lambda L] = T L + 2 \lambda L + \frac{\lambda^3}{12} C.

2. The **affine** or current Lie conformal algebra `L(\mathfrak{g})`
   associated to a finite dimensional Lie algebra `\mathfrak{g}` with
   non-degenerate, invariant `R`-bilinear form `(,)` is given as a central
   extension of the free
   `R[T]` module generated by `\mathfrak{g}` by a central element `K`. The
   `\lambda`-bracket of generators is given by

   .. MATH::

        [a_\lambda b] = [a,b] + \lambda (a,b) K, \qquad a,b \in \mathfrak{g}

3. The **Weyl** Lie conformal algebra, or `\beta-\gamma` system is
   given as the central extension of a free `R[T]` module with two
   generators `\beta` and `\gamma`, by a central element `K`.
   The only non-trivial brackets among generators are

   .. MATH::

        [\beta_\lambda \gamma] = - [\gamma_\lambda \beta] = K

4. The **Neveu-Schwarz** super Lie conformal algebra is a super Lie
   conformal algebra which is an extension of the Virasoro Lie conformal
   algebra. It consists of a Virasoro generator `L` as in example 1 above
   and an *odd* generator `G`. The remaining brackets are given by:

   .. MATH::

        [L_\lambda G] = \left( T + \frac{3}{2} \lambda \right) G \qquad
        [G_\lambda G] = 2 L + \frac{\lambda^2}{3} C

.. SEEALSO::

    :mod:`sage.algebras.lie_conformal_algebras.examples`

The base class for all Lie conformal algebras is
:class:`LieConformalAlgebra`.
All subclasses are called through its method ``__classcall_private__``.
This class provides no functionality besides calling the appropriate
constructor.

We provide some convenience classes to define named Lie conformal
algebras. See
:mod:`sage.algebras.lie_conformal_algebras.examples`.

EXAMPLES:

- We construct the Virasoro Lie conformal algebra, its universal
  enveloping vertex algebra and lift some elements::

    sage: Vir = lie_conformal_algebras.Virasoro(QQ)
    sage: Vir.inject_variables()
    Defining L, C
    sage: L.bracket(L)
    {0: TL, 1: 2*L, 3: 1/2*C}

- We construct the Current algebra for `\mathfrak{sl}_2`::

    sage: R = lie_conformal_algebras.Affine(QQ, 'A1', names = ('e', 'h', 'f'))
    sage: R.gens()
    (e, h, f, K)
    sage: R.inject_variables()
    Defining e, h, f, K
    sage: e.bracket(f.T())
    {0: Th, 1: h, 2: 2*K}
    sage: e.T(3)
    6*T^(3)e

- We construct the `\beta-\gamma` system by directly giving the
  `\lambda`-brackets of the generators::

    sage: betagamma_dict = {('b','a'):{0:{('K',0):1}}}
    sage: V = LieConformalAlgebra(QQ, betagamma_dict, names=('a','b'), weights=(1,0), central_elements=('K',))
    sage: V.category()
    Category of H-graded finitely generated Lie conformal algebras with basis over Rational Field
    sage: V.inject_variables()
    Defining a, b, K
    sage: a.bracket(b)
    {0: -K}

AUTHORS:

- Reimundo Heluani (2019-08-09): Initial implementation.
"""


#******************************************************************************
#       Copyright (C) 2019 Reimundo Heluani <heluani@potuz.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.unique_representation import UniqueRepresentation
from sage.sets.family import Family
from sage.categories.commutative_rings import CommutativeRings
from sage.structure.parent import Parent


class LieConformalAlgebra(UniqueRepresentation, Parent):
    r"""
    Lie Conformal Algebras base class and factory.

    INPUT:

    - ``R`` -- a commutative ring (default: ``None``); the base
      ring of this Lie conformal algebra. Behaviour is undefined
      if it is not a field of characteristic zero.

    - ``arg0`` -- dictionary (default: ``None``);
      a dictionary containing the `\lambda` brackets of the
      generators of this Lie conformal algebra. The keys of this
      dictionary are pairs of either names or indices of the
      generators and the values are themselves dictionaries. For a
      pair of generators ``'a'`` and ``'b'``, the value of
      ``arg0[('a','b')]`` is a dictionary whose keys are positive
      integer numbers and the corresponding value for the
      key ``j`` is a dictionary itself representing the `j`-th product
      `a_{(j)}b`. Thus, for a positive integer number `j`, the
      value of ``arg0[('a','b')][j]`` is a dictionary whose entries
      are pairs ``('c',n)`` where ``'c'`` is the name of a generator
      and ``n`` is a positive number. The value for this key is the
      coefficient of `\frac{T^{n}}{n!} c` in `a_{(j)}b`. For
      example the ``arg0`` for the *Virasoro* Lie conformal algebra
      is::

            {('L','L'):{0:{('L',1):1}, 1:{('L',0):2}, 3:{('C',0):1/2}}}


      Do not include central elements as keys in this dictionary. Also,
      if the key ``('a','b')`` is present, there is no need to include
      ``('b','a')`` as it is defined by skew-symmetry. Any missing
      pair (besides the ones defined by skew-symmetry) is assumed
      to have vanishing `\lambda`-bracket.

    - ``names`` -- tuple of strings (default: ``None``); the list of
      names for generators of this Lie conformal algebra. Do not
      include central elements in this list.

    - ``central_elements`` -- tuple of strings (default: ``None``);
      a list of names for central elements of this Lie conformal algebra

    - ``index_set`` -- enumerated set (default: ``None``); an
      indexing set for the generators of this Lie conformal algebra.
      Do not include central elements in this list.

    - ``weights`` -- tuple of nonnegative rational numbers
      (default: ``None``); a list of degrees for this Lie
      conformal algebra.
      The returned Lie conformal algebra is H-Graded. This tuple
      needs to have the same cardinality as ``index_set`` or
      ``names``. Central elements are assumed to have weight `0`.

    - ``parity`` -- tuple of `0` or `1` (default: tuple of `0`);
      if this is a super Lie conformal algebra, this tuple
      specifies the parity of each of the non-central generators of
      this Lie conformal algebra. Central elements are assumed to
      be even. Notice that if this tuple is present, the category
      of this Lie conformal algebra is set to be a subcategory of
      ``LieConformalAlgebras(R).Super()``, even if all generators
      are even.

    - ``category`` -- the category that this Lie conformal algebra
      belongs to

    In addition we accept the following keywords:

    - ``graded`` -- boolean (default: ``False``);
      if ``True``, the returned algebra is H-Graded.
      If ``weights`` is not specified, all non-central generators
      are assigned degree `1`. This keyword is ignored if
      ``weights`` is specified

    - ``super`` -- boolean (default: ``False``);
      if ``True``, the returned algebra is a super
      Lie conformal algebra even if all generators are even.
      If ``parity`` is not specified, all generators are
      assigned even parity. This keyword is ignored if
      ``parity`` is specified.

    .. Note::

        Any remaining keyword is currently passed to
        :class:`CombinatorialFreeModule<sage.combinat.free_module.CombinatorialFreeModule>`.

    EXAMPLES:

    We construct the `\beta-\gamma` system or *Weyl* Lie conformal
    algebra::

        sage: betagamma_dict = {('b','a'):{0:{('K',0):1}}}
        sage: V = LieConformalAlgebra(QQbar, betagamma_dict, names=('a','b'), weights=(1,0), central_elements=('K',))
        sage: V.category()
        Category of H-graded finitely generated Lie conformal algebras with basis over Algebraic Field
        sage: V.inject_variables()
        Defining a, b, K
        sage: a.bracket(b)
        {0: -K}

    We construct the current algebra for `\mathfrak{sl}_2`::

        sage: sl2dict = {('e','f'):{0:{('h',0):1}, 1:{('K',0):1}}, ('e','h'):{0:{('e',0):-2}}, ('f','h'):{0:{('f',0):2}}, ('h', 'h'):{1:{('K', 0):2}}}
        sage: V = LieConformalAlgebra(QQ, sl2dict, names=('e', 'h', 'f'), central_elements=('K',), graded=True)
        sage: V.inject_variables()
        Defining e, h, f, K
        sage: e.bracket(f)
        {0: h, 1: K}
        sage: h.bracket(e)
        {0: 2*e}
        sage: e.bracket(f.T())
        {0: Th, 1: h, 2: 2*K}
        sage: V.category()
        Category of H-graded finitely generated Lie conformal algebras with basis over Rational Field
        sage: e.degree()
        1

    .. TODO::

        This class checks that the provided dictionary is consistent
        with skew-symmetry. It does not check that it is consistent
        with the Jacobi identity.

    .. SEEALSO::

        :mod:`sage.algebras.lie_conformal_algebras.graded_lie_conformal_algebra`
    """
    @staticmethod
    def __classcall_private__(cls, R=None, arg0=None, index_set=None,
                              central_elements=None, category=None,
                              prefix=None, names=None, latex_names=None,
                              parity=None, weights=None, **kwds):
        """
        Lie conformal algebra factory.

        EXAMPLES::

            sage: betagamma_dict = {('b','a'):{0:{('K',0):1}}}
            sage: V = LieConformalAlgebra(QQ, betagamma_dict, names=('a','b'), weights=(1,0), central_elements=('K',))
            sage: type(V)
            <class 'sage.algebras.lie_conformal_algebras.graded_lie_conformal_algebra.GradedLieConformalAlgebra_with_category'>
        """
        if R not in CommutativeRings():
            raise ValueError(f"arg0 must be a commutative ring got {R}")

        # This is the only exposed class so we clean keywords here
        known_keywords = ['category', 'prefix', 'bracket', 'latex_bracket',
                          'string_quotes', 'sorting_key', 'graded', 'super']
        for key in kwds:
            if key not in known_keywords:
                raise ValueError("got an unexpected keyword argument '%s'" % key)

        if isinstance(arg0, dict) and arg0:
            graded = kwds.pop("graded", False)
            if weights is not None or graded:
                from .graded_lie_conformal_algebra import \
                                                    GradedLieConformalAlgebra
                return GradedLieConformalAlgebra(R, Family(arg0),
                    index_set=index_set, central_elements=central_elements,
                    category=category, prefix=prefix, names=names,
                    latex_names=latex_names, parity=parity, weights=weights,
                    **kwds)
            else:
                from .lie_conformal_algebra_with_structure_coefs import \
                        LieConformalAlgebraWithStructureCoefficients
                return LieConformalAlgebraWithStructureCoefficients(R,
                       Family(arg0), index_set=index_set,
                       central_elements=central_elements, category=category,
                       prefix=prefix, names=names, latex_names=latex_names,
                       parity=parity, **kwds)
        raise NotImplementedError("not implemented")



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Star-crystal structure on `B(\infty)`

AUTHORS:

- Ben Salisbury: initial version
- Travis Scrimshaw: initial version
"""

#*****************************************************************************
#       Copyright (C) 2016 Ben Salisbury <ben DOT salisbury AT cmich DOT edu>
#                          Travis Scrimshaw <tscrimsh AT umn DOT edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.element_wrapper import ElementWrapper
from sage.categories.highest_weight_crystals import HighestWeightCrystals
from sage.combinat.crystals.elementary_crystals import ElementaryCrystal


class StarCrystal(UniqueRepresentation, Parent):
    r"""
    The star-crystal or `*`-crystal version of a highest weight crystal.

    The `*`-crystal structure on `B(\infty)` is the structure induced by
    the algebra antiautomorphism `* \colon U_q(\mathfrak{g}) \longrightarrow
    U_q(\mathfrak{g})` that stabilizes the negative half `U_q^-(\mathfrak{g})`.
    It is defined by

    .. MATH::

        E_i^* = E_i , \ \ \
        F_i^* = F_i , \ \ \
        q^* = q, \ \ \
        (q^h)^* = q^{-h},

    where `E_i` and `F_i` are the Chevalley generators of `U_q(\mathfrak{g})`
    and `h` is an element of the Cartan subalgebra.

    The induced operation on the crystal `B(\infty)` is called the
    *Kashiwara involution*.  Its implementation here is based on the
    recursive algorithm from Theorem 2.2.1 of [Ka1993]_, which states
    that for any `i \in I` there is a unique strict crystal embedding

    .. MATH::

        \Psi_i\colon B(\infty) \longrightarrow B_i \otimes B(\infty)

    such that

    - `u_{\infty} \mapsto b_i(0) \otimes u_{\infty}`, where `u_{\infty}`
      is the highest weight vector in `B(\infty)`;

    - if `\Psi_i(b) = f_i^mb_i(0) \otimes b_0`, then
      `\Psi_i(f_i^*b) = f_i^{m+1}b_i(0) \otimes b_0`
      and `\varepsilon_i(b^*) = m`;

    - the image of `\Psi_i` is `\{f_i^mb_i(0)\otimes b :
      \varepsilon_i(b^*) = 0, \ m\ge 0\}`.

    Here, `B_i` is the `i`-th elementary crystal.  See
    :class:`~sage.combinat.crystals.elementary_crystals.ElementaryCrystal`
    for more information.

    INPUT:

    - ``Binf`` -- a crystal from
      :class:`~sage.combinat.crystals.catalog_infinity_crystals`

    EXAMPLES::

        sage: B = crystals.infinity.Tableaux(['A',2])
        sage: Bstar = crystals.infinity.Star(B)
        sage: mg = Bstar.highest_weight_vector()
        sage: mg
        [[1, 1], [2]]
        sage: mg.f_string([1,2,1,2,2])
        [[1, 1, 1, 1, 1, 2, 2], [2, 3, 3, 3]]
    """

    def __init__(self, Binf):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: B = crystals.infinity.Tableaux(['A',2])
            sage: Bstar = crystals.infinity.Star(B)
            sage: TestSuite(Bstar).run(max_runs=40)
            sage: TestSuite(Bstar).run(max_runs=1000) # long time
        """
        self._Binf = Binf
        self._cartan_type = Binf.cartan_type()
        Parent.__init__(self, category=HighestWeightCrystals().Infinite())
        self.module_generators = (self(self._Binf.module_generators[0]),)
        t0 = Binf.highest_weight_vector()
        B = {i: ElementaryCrystal(Binf.cartan_type(),i) for i in self.index_set()}
        self._tens = {i: B[i].tensor(Binf) for i in self.index_set()}
        gens = {i: self._tens[i](B[i](0), t0) for i in self.index_set()}
        self._embedding = {i: Binf.crystal_morphism({t0: gens[i]}) for i in self.index_set()}
        self._pullback = {i: self._tens[i].crystal_morphism({gens[i]: t0}) for i in self.index_set()}

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: Y = crystals.infinity.GeneralizedYoungWalls(3)
            sage: Ystar = crystals.infinity.Star(Y)
            sage: Ystar
            Star-crystal version of Crystal of generalized Young walls of type ['A', 3, 1]
        """
        return "Star-crystal version of %s" % self._Binf

    class Element(ElementWrapper):

        def e(self, i):
            r"""
            Return the action of `e_i^*` on ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: RC = crystals.infinity.RiggedConfigurations(['E',6,1])
                sage: RCstar = crystals.infinity.Star(RC)
                sage: nuJ = RCstar.module_generators[0].f_string([0,4,6,1,2])
                sage: ascii_art(nuJ.e(1))
                -1[ ]-1   (/)   0[ ]1   (/)   -1[ ]-1   (/)   -2[ ]-1

                sage: M = crystals.infinity.NakajimaMonomials(['B',2,1])
                sage: Mstar = crystals.infinity.Star(M)
                sage: m = Mstar.module_generators[0].f_string([0,1,2,2,1,0])
                sage: m.e(1)
                Y(0,0)^-1 Y(0,2)^-1 Y(1,1) Y(1,2)^-1 Y(2,1)^2
            """
            P = self.parent()
            image = P._embedding[i](self.value)
            if image[0].e(i)._m > 0:
                return None
            return P(P._pullback[i]( P._tens[i](image[0].e(i),image[1]) ))

        def f(self, i):
            r"""
            Return the action of `f_i^*` on ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: T = crystals.infinity.Tableaux("G2")
                sage: Tstar = crystals.infinity.Star(T)
                sage: t = Tstar.module_generators[0].f_string([1,2,1,1,2])
                sage: t
                [[1, 1, 1, 2, 0], [2, 3]]

                sage: M = crystals.infinity.NakajimaMonomials(['B',2,1])
                sage: Mstar = crystals.infinity.Star(M)
                sage: m = Mstar.module_generators[0].f_string([0,1,2,2,1,0])
                sage: m
                Y(0,0)^-1 Y(0,2)^-1 Y(1,0)^-1 Y(1,2)^-1 Y(2,0)^2 Y(2,1)^2
            """
            P = self.parent()
            image = P._embedding[i](self.value)
            return P(P._pullback[i]( P._tens[i](image[0].f(i),image[1]) ))

        def weight(self):
            r"""
            Return the weight of ``self``.

            EXAMPLES::

                sage: RC = crystals.infinity.RiggedConfigurations(['E',6,1])
                sage: RCstar = crystals.infinity.Star(RC)
                sage: nuJ = RCstar.module_generators[0].f_string([0,4,6,1,2])
                sage: nuJ.weight()
                -Lambda[0] - 2*Lambda[1] + 2*Lambda[3] - Lambda[4]
                 + 2*Lambda[5] - 2*Lambda[6] - delta
            """
            return self.value.weight()

        def epsilon(self, i):
            r"""
            Return `\varepsilon_i^*` of ``self``.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: Y = crystals.infinity.GeneralizedYoungWalls(3)
                sage: Ystar = crystals.infinity.Star(Y)
                sage: y = Ystar.module_generators[0].f_string([0,1,3,2,1,0])
                sage: [y.epsilon(i) for i in y.index_set()]
                [1, 0, 1, 0]

                sage: RC = crystals.infinity.RiggedConfigurations(['E',6,1])
                sage: RCstar = crystals.infinity.Star(RC)
                sage: nuJ = RCstar.module_generators[0].f_string([0,4,6,1,2])
                sage: [nuJ.epsilon(i) for i in nuJ.index_set()]
                [0, 1, 1, 0, 0, 0, 1]
            """
            ep = -1
            while self is not None:
                ep += 1
                self = self.e(i)
            return ep

        def phi(self, i):
            r"""
            Return `\varphi_i^*` of ``self``.

            For `b \in B(\infty)`,

            .. MATH::

                \varphi_i^*(b) = \varepsilon_i^*(b) + \langle h_i,
                \mathrm{wt}(b) \rangle,

            where `h_i` is a simple coroot.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: T = crystals.infinity.Tableaux("A2")
                sage: Tstar = crystals.infinity.Star(T)
                sage: t = Tstar.module_generators[0].f_string([1,2,1,1,2])
                sage: [t.phi(i) for i in t.index_set()]
                [-3, 1]

                sage: M = crystals.infinity.NakajimaMonomials(['B',2,1])
                sage: Mstar = crystals.infinity.Star(M)
                sage: m = Mstar.module_generators[0].f_string([0,1,2,2,1,0])
                sage: [m.phi(i) for i in m.index_set()]
                [-1, -1, 4]
            """
            P = self.parent().weight_lattice_realization()
            ac = P.simple_coroot(i)
            return P(self.weight()).scalar(ac) + self.epsilon(i)

        def jump(self, i):
            r"""
            Return the `i`-jump of ``self``.

            For `b \in B(\infty)`,

            .. MATH::

                \operatorname{jump}_i(b) = \varepsilon_i(b) + \varepsilon_i^*(b)
                + \langle h_i, \mathrm{wt}(b) \rangle,

            where `h_i` is a simple coroot.

            INPUT:

            - ``i`` -- an element of the index set

            EXAMPLES::

                sage: RC = crystals.infinity.RiggedConfigurations("D4")
                sage: RCstar = crystals.infinity.Star(RC)
                sage: nu0star = RCstar.module_generators[0]
                sage: nustar = nu0star.f_string([2,1,3,4,2])
                sage: [nustar.jump(i) for i in RC.index_set()]
                [0, 1, 0, 0]
                sage: nustar = nu0star.f_string([2,1,3,4,2,2,1,3,2]) # long time
                sage: [nustar.jump(i) for i in RC.index_set()] # long time
                [1, 0, 1, 2]
            """
            P = self.parent().weight_lattice_realization()
            ac = P.simple_coroot(i)
            return P(self.value.weight()).scalar(ac) + self.epsilon(i) + self.value.epsilon(i)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
r"""
Family Games America's Quantumino solver

This module allows to solve the `Quantumino puzzle
<http://familygamesamerica.com/mainsite/consumers/productview.php?pro_id=274&search=quantumino>`_
made by Family Games America (see also `this video
<http://www.youtube.com/watch?v=jX_VKzakZi8>`_ on Youtube). This puzzle was
left at the dinner room of the Laboratoire de Combinatoire Informatique
Mathématique in Montreal by Franco Saliola during winter 2011.

The solution uses the dancing links code which is in Sage and is based on
the more general code available in the module :mod:`sage.combinat.tiling`.
Dancing links were originally introduced by Donald Knuth in 2000
(:arxiv:`cs/0011047`). In particular,
Knuth used dancing links to solve tilings of a region by 2D pentaminos.
Here we extend the method for 3D pentaminos.

This module defines two classes :

- :class:`sage.games.quantumino.QuantuminoState` class, to represent a
  state of the Quantumino game, i.e. a solution or a partial solution.

- :class:`sage.games.quantumino.QuantuminoSolver` class, to find, enumerate
  and count the number of solutions of the Quantumino game where one of the
  piece is put aside.

AUTHOR:

- Sébastien Labbé, April 28th, 2011

DESCRIPTION (from [1]):

    "
    Pentamino games have been taken to a whole different level; a 3-D
    level, with this colorful creation! Using the original pentamino
    arrangements of 5 connected squares which date from 1907, players are
    encouraged to "think inside the box" as they try to fit 16 of the 17
    3-D pentamino pieces inside the playing perimeters. Remove a different
    piece each time you play for an entirely new challenge! Thousands of
    solutions to be found!
    Quantumino hands-on educational tool where players learn how shapes
    can be transformed or arranged into predefined shapes and spaces.
    Includes:
    1 wooden frame, 17 wooden blocks, instruction booklet.
    Age: 8+
    "

EXAMPLES:

Here are the 17 wooden blocks of the Quantumino puzzle numbered from 0 to 16 in
the following 3d picture. They will show up in 3D in your default (=Jmol)
viewer::

    sage: from sage.games.quantumino import show_pentaminos
    sage: show_pentaminos()                                                             # needs sage.plot
    Graphics3d Object

To solve the puzzle where the pentamino numbered 12 is put aside::

    sage: from sage.games.quantumino import QuantuminoSolver
    sage: s = next(QuantuminoSolver(12).solve())          # long time (10 s)
    sage: s                                               # long time (<1s)
    Quantumino state where the following pentamino is put aside :
    Polyomino: [(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1), (2, 1, 1)], Color: blue
    sage: s.show3d()                            # long time (<1s)                       # needs sage.plot
    Graphics3d Object

To remove the frame::

    sage: s.show3d().show(frame=False)          # long time (<1s)                       # needs sage.plot

To solve the puzzle where the pentamino numbered 7 is put aside::

    sage: s = next(QuantuminoSolver(7).solve())           # long time (10 s)
    sage: s                                               # long time (<1s)
    Quantumino state where the following pentamino is put aside :
    Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (0, 2, 1), (1, 0, 0)], Color: orange
    sage: s.show3d()                            # long time (<1s)                       # needs sage.plot
    Graphics3d Object

The solution is iterable. This may be used to explicitly list the positions of each
pentamino::

    sage: for p in s: p                                   # long time (<1s)
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...
    Polyomino: [(...), (...), (...), (...), (...)], Color: ...

To get all the solutions, use the iterator returned by the ``solve``
method. Note that finding the first solution is the most time consuming
because it needs to create the complete data to describe the problem::

    sage: it = QuantuminoSolver(7).solve()
    sage: next(it)                                     # not tested (10s)
    Quantumino state where the following pentamino is put aside :
    Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (0, 2, 1), (1, 0, 0)], Color: orange
    sage: next(it)                                     # not tested (0.001s)
    Quantumino state where the following pentamino is put aside :
    Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (0, 2, 1), (1, 0, 0)], Color: orange
    sage: next(it)                                     # not tested (0.001s)
    Quantumino state where the following pentamino is put aside :
    Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (0, 2, 1), (1, 0, 0)], Color: orange

To get the solution inside other boxes::

    sage: s = next(QuantuminoSolver(7, box=(4,4,5)).solve())        # not tested (2s)
    sage: s.show3d()                            # not tested (<1s)                      # needs sage.plot

::

    sage: s = next(QuantuminoSolver(7, box=(2,2,20)).solve())       # not tested (1s)
    sage: s.show3d()                            # not tested (<1s)                      # needs sage.plot

If there are no solution, a StopIteration error is raised::

    sage: next(QuantuminoSolver(7, box=(3,3,3)).solve())
    Traceback (most recent call last):
    ...
    StopIteration

The implementation allows a lot of introspection. From the
:class:`~sage.combinat.tiling.TilingSolver` object,
it is possible to retrieve the rows that are passed to the DLX
solver and count them. It is also possible to get an instance of the DLX
solver to play with it::

    sage: q = QuantuminoSolver(0)
    sage: T = q.tiling_solver()
    sage: T
    Tiling solver of 16 pieces into a box of size 80
    Rotation allowed: True
    Reflection allowed: False
    Reusing pieces allowed: False
    sage: rows = T.rows()                            # not tested (10 s)
    sage: len(rows)                                  # not tested (but fast)
    5484
    sage: x = T.dlx_solver()                         # long time (10 s)
    sage: x                                          # long time (fast)
    Dancing links solver for 96 columns and 5484 rows

TESTS:

We check that all pentaminos are equal to their canonical translate::

    sage: from sage.games.quantumino import pentaminos
    sage: all(p == p.canonical() for p in pentaminos)
    True

REFERENCES:

- [1] `Family Games America's Quantumino
  <http://familygamesamerica.com/mainsite/consumers/productview.php?pro_id=274&search=quantumino>`_
- [2] `Quantumino - How to Play <http://www.youtube.com/watch?v=jX_VKzakZi8>`_ on Youtube
- [3] Knuth, Donald (2000). *Dancing links*. :arxiv:`cs/0011047`.
"""
# ****************************************************************************
#       Copyright (C) 2011 Sebastien Labbe <slabqc@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.sage_object import SageObject
from sage.misc.lazy_import import lazy_import
lazy_import("sage.plot.all", "Graphics")
lazy_import("sage.plot.plot3d.platonic", "cube")
lazy_import("sage.plot.plot3d.shapes2", "text3d")
from sage.modules.free_module_element import vector
from sage.combinat.tiling import Polyomino, TilingSolver

################################################
# Example:  The family games america: Quantumino
################################################
pentaminos = []
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,2,0), (1,1,1)], color='deeppink'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,0,1), (2,0,0), (2,1,0)], color='deeppink'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,2,0), (0,0,1)], color='green'))
pentaminos.append(Polyomino([(0,0,0), (0,1,0), (0,2,0), (1,0,0), (1,0,1)], color='green'))
pentaminos.append(Polyomino([(0,1,0), (1,0,1), (1,1,0), (1,1,1), (1,2,0)], color='red'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,0,1), (2,0,1)], color='red'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,2,0), (1,2,1)], color='orange'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (0,1,0), (0,2,0), (0,2,1)], color='orange'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (0,1,0), (1,1,0), (0,0,1)], color='yellow'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,1,1), (0,0,1)], color='yellow'))
pentaminos.append(Polyomino([(0,0,0), (0,1,0), (1,1,0), (0,2,0), (1,1,1)], color='midnightblue'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,0,1), (1,2,0)], color='darkblue'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (1,1,1), (2,1,1)], color='blue'))
pentaminos.append(Polyomino([(0,0,0), (0,1,0), (1,1,0), (1,1,1), (1,2,1)], color='blue'))
pentaminos.append(Polyomino([(0,0,0), (1,0,0), (1,1,0), (2,1,0), (2,1,1)], color='purple'))
pentaminos.append(Polyomino([(0,0,0), (0,1,0), (1,1,0), (1,2,0), (1,2,1)], color='purple'))
pentaminos.append(Polyomino([(0,1,0), (1,0,0), (1,1,0), (1,1,1), (1,2,0)], color='gray'))


def show_pentaminos(box=(5, 8, 2)):
    r"""
    Show the 17 3-D pentaminos included in the game and the `5 \times 8
    \times 2` box where 16 of them must fit.

    INPUT:

    - ``box`` -- tuple of size three (default: ``(5,8,2)``),
      size of the box

    OUTPUT: 3D Graphic object

    EXAMPLES::

        sage: from sage.games.quantumino import show_pentaminos
        sage: show_pentaminos()    # not tested (1s)

    To remove the frame do::

        sage: show_pentaminos().show(frame=False)  # not tested (1s)
    """
    G = Graphics()
    for i, p in enumerate(pentaminos):
        x = 4 * (i % 4)
        y = 4 * (i // 4)
        q = p + (x, y, 0)
        G += q.show3d()
        G += text3d(str(i), (x, y, 2))
    G += cube(color='gray',opacity=0.5).scale(box).translate((17, 6, 0))

    # hack to set the aspect ratio to 1
    a, b = G.bounding_box()
    a, b = map(vector, (a, b))
    G.frame_aspect_ratio(tuple(b - a))

    return G

##############################
# Class QuantuminoState
##############################


class QuantuminoState(SageObject):
    r"""
    A state of the Quantumino puzzle.

    Used to represent a solution or a partial solution of the Quantumino
    puzzle.

    INPUT:

    - ``pentos`` -- list of 16 3d pentamino representing the (partial)
      solution
    - ``aside`` -- 3d polyomino, the unused 3D pentamino
    - ``box`` -- tuple of size three (default: ``(5,8,2)``),
      size of the box

    EXAMPLES::

        sage: from sage.games.quantumino import pentaminos, QuantuminoState
        sage: p = pentaminos[0]
        sage: q = pentaminos[5]
        sage: r = pentaminos[11]
        sage: S = QuantuminoState([p,q], r)
        sage: S
        Quantumino state where the following pentamino is put aside :
        Polyomino: [(0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0), (1, 2, 0)], Color: darkblue

    ::

        sage: from sage.games.quantumino import QuantuminoSolver
        sage: next(QuantuminoSolver(3).solve())      # not tested (1.5s)
        Quantumino state where the following pentamino is put aside :
        Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (1, 0, 0), (1, 0, 1)], Color: green
    """
    def __init__(self, pentos, aside, box=(5, 8, 2)):
        r"""
        EXAMPLES::

            sage: from sage.games.quantumino import pentaminos, QuantuminoState
            sage: p = pentaminos[0]
            sage: q = pentaminos[5]
            sage: QuantuminoState([p], q)
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0), (2, 0, 1)], Color: red
        """
        assert all(isinstance(p, Polyomino) for p in pentos), "pentos must be an iterable of Polyomino"
        assert isinstance(aside, Polyomino), "aside must be a Polyomino"
        self._pentos = pentos
        self._aside = aside
        self._box = box

    def __repr__(self):
        r"""
        EXAMPLES::

            sage: from sage.games.quantumino import pentaminos, QuantuminoState
            sage: p = pentaminos[0]
            sage: q = pentaminos[5]
            sage: QuantuminoState([p], q)
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0), (2, 0, 1)], Color: red
        """
        return "Quantumino state where the following pentamino is put aside :\n%s" % self._aside

    def __iter__(self):
        r"""
        EXAMPLES::

            sage: from sage.games.quantumino import pentaminos, QuantuminoState
            sage: p = pentaminos[0]
            sage: q = pentaminos[5]
            sage: r = pentaminos[11]
            sage: S = QuantuminoState([p,q], r)
            sage: for a in S: a
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1), (1, 2, 0)], Color: deeppink
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0), (2, 0, 1)], Color: red
        """
        return iter(self._pentos)

    def list(self):
        r"""
        Return the list of 3d polyomino making the solution.

        EXAMPLES::

            sage: from sage.games.quantumino import pentaminos, QuantuminoState
            sage: p = pentaminos[0]
            sage: q = pentaminos[5]
            sage: r = pentaminos[11]
            sage: S = QuantuminoState([p,q], r)
            sage: L = S.list()
            sage: L[0]
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1), (1, 2, 0)], Color: deeppink
            sage: L[1]
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0), (2, 0, 1)], Color: red
        """
        return list(self)

    def show3d(self, size=0.85):
        r"""
        Return the solution as a 3D Graphic object.

        OUTPUT: 3D Graphic Object

        EXAMPLES::

            sage: from sage.games.quantumino import QuantuminoSolver
            sage: s = next(QuantuminoSolver(0).solve())    # not tested (1.5s)
            sage: G = s.show3d()                            # not tested (<1s)
            sage: type(G)                                   # not tested
            <class 'sage.plot.plot3d.base.Graphics3dGroup'>

        To remove the frame::

            sage: G.show(frame=False) # not tested

        To see the solution with Tachyon viewer::

            sage: G.show(viewer='tachyon', frame=False) # not tested
        """
        G = Graphics()
        for p in self:
            G += p.show3d(size=size)
        aside_pento = self._aside.canonical() + (2,-4,0)
        G += aside_pento.show3d(size=size)

        # the box to fill
        half_box = tuple(a/2 for a in self._box)
        b = cube(color='gray',opacity=0.2).scale(self._box).translate(half_box)
        b = b.translate((0, -.5, -.5))
        G += b

        # hack to set the aspect ratio to 1
        a,b = G.bounding_box()
        a,b = map(vector, (a,b))
        G.frame_aspect_ratio(tuple(b-a))

        return G

##############################
# Class QuantuminoSolver
##############################


class QuantuminoSolver(SageObject):
    r"""
    Return the Quantumino solver for the given box where one of the
    pentamino is put aside.

    INPUT:

    - ``aside`` -- integer, from 0 to 16, the aside pentamino
    - ``box`` -- tuple of size three (default: ``(5,8,2)``),
      size of the box

    EXAMPLES::

        sage: from sage.games.quantumino import QuantuminoSolver
        sage: QuantuminoSolver(9)
        Quantumino solver for the box (5, 8, 2)
        Aside pentamino number: 9
        sage: QuantuminoSolver(12, box=(5,4,4))
        Quantumino solver for the box (5, 4, 4)
        Aside pentamino number: 12
    """
    def __init__(self, aside, box=(5, 8, 2)):
        r"""
        Constructor.

        EXAMPLES::

            sage: from sage.games.quantumino import QuantuminoSolver
            sage: QuantuminoSolver(9)
            Quantumino solver for the box (5, 8, 2)
            Aside pentamino number: 9
        """
        if not (0 <= aside < 17):
            raise ValueError(f"aside (={aside}) must be between 0 and 16")
        self._aside = aside
        self._box = box

    def __repr__(self):
        r"""
        String representation.

        EXAMPLES::

            sage: from sage.games.quantumino import QuantuminoSolver
            sage: QuantuminoSolver(0)
            Quantumino solver for the box (5, 8, 2)
            Aside pentamino number: 0
        """
        s = "Quantumino solver for the box %s\n" % (self._box, )
        s += "Aside pentamino number: %s" % self._aside
        return s

    def tiling_solver(self):
        r"""
        Return the Tiling solver of the Quantumino Game where one of the
        pentamino is put aside.

        EXAMPLES::

            sage: from sage.games.quantumino import QuantuminoSolver
            sage: QuantuminoSolver(0).tiling_solver()
            Tiling solver of 16 pieces into a box of size 80
            Rotation allowed: True
            Reflection allowed: False
            Reusing pieces allowed: False
            sage: QuantuminoSolver(14).tiling_solver()
            Tiling solver of 16 pieces into a box of size 80
            Rotation allowed: True
            Reflection allowed: False
            Reusing pieces allowed: False
            sage: QuantuminoSolver(14, box=(5,4,4)).tiling_solver()
            Tiling solver of 16 pieces into a box of size 80
            Rotation allowed: True
            Reflection allowed: False
            Reusing pieces allowed: False
        """
        pieces = pentaminos[:self._aside] + pentaminos[self._aside+1:]
        return TilingSolver(pieces, box=self._box)

    def solve(self, partial=None):
        r"""
        Return an iterator over the solutions where one of the pentamino is
        put aside.

        INPUT:

        - ``partial`` -- string (default: ``None``), whether to
          include partial (incomplete) solutions. It can be one of the
          following:

          - ``None`` -- include only complete solution
          - ``'common'`` -- common part between two consecutive solutions
          - ``'incremental'`` -- one piece change at a time

        OUTPUT: iterator of :class:`QuantuminoState`

        EXAMPLES:

        Get one solution::

            sage: from sage.games.quantumino import QuantuminoSolver
            sage: s = next(QuantuminoSolver(8).solve())          # long time (9s)
            sage: s                                              # long time (fast)
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 1, 0)], Color: yellow
            sage: s.show3d()                                     # long time (< 1s)
            Graphics3d Object

        The explicit solution::

            sage: for p in s: p                                  # long time (fast)
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...
            Polyomino: [(...), (...), (...), (...), (...)], Color: ...


        Enumerate the solutions::

            sage: it = QuantuminoSolver(0).solve()
            sage: next(it)                                          # not tested
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1), (1, 2, 0)], Color: deeppink
            sage: next(it)                                          # not tested
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1), (1, 2, 0)], Color: deeppink

        With the partial solutions included, one can see the evolution
        between consecutive solutions (an animation would be better)::

            sage: it = QuantuminoSolver(0).solve(partial='common')
            sage: next(it).show3d()               # not tested (2s)
            sage: next(it).show3d()               # not tested (< 1s)
            sage: next(it).show3d()               # not tested (< 1s)

        Generalizations of the game inside different boxes::

            sage: next(QuantuminoSolver(7, (4,4,5)).solve())       # long time (2s)
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (0, 2, 1), (1, 0, 0)], Color: orange
            sage: next(QuantuminoSolver(7, (2,2,20)).solve())      # long time (1s)
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (0, 2, 1), (1, 0, 0)], Color: orange
            sage: next(QuantuminoSolver(3, (2,2,20)).solve())      # long time (1s)
            Quantumino state where the following pentamino is put aside :
            Polyomino: [(0, 0, 0), (0, 1, 0), (0, 2, 0), (1, 0, 0), (1, 0, 1)], Color: green

        If the volume of the box is not 80, there is no solution::

            sage: next(QuantuminoSolver(7, box=(3,3,9)).solve())
            Traceback (most recent call last):
            ...
            StopIteration

        If the box is too small, there is no solution::

            sage: next(QuantuminoSolver(4, box=(40,2,1)).solve())
            Traceback (most recent call last):
            ...
            StopIteration
        """
        T = self.tiling_solver()
        aside = pentaminos[self._aside]
        for pentos in T.solve(partial=partial):
            yield QuantuminoState(pentos, aside, self._box)

    def number_of_solutions(self):
        r"""
        Return the number of solutions.

        OUTPUT: integer

        EXAMPLES::

            sage: from sage.games.quantumino import QuantuminoSolver
            sage: QuantuminoSolver(4, box=(3,2,2)).number_of_solutions()
            0

        This computation takes several days::

            sage: QuantuminoSolver(0).number_of_solutions()                # not tested
            ??? hundreds of millions ???
        """
        return self.tiling_solver().number_of_solutions()



# sage_setup: distribution = sagemath-combinat
r"""
Ribbons
"""
#*****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from .ribbon_shaped_tableau import RibbonShapedTableau, StandardRibbonShapedTableaux



# sage_setup: distribution = sagemath-combinat
"""
Set species
"""
# ****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from .species import GenericCombinatorialSpecies
from sage.combinat.species.structure import GenericSpeciesStructure
from sage.combinat.species.misc import accept_size
from sage.structure.unique_representation import UniqueRepresentation
from sage.arith.misc import factorial


class SetSpeciesStructure(GenericSpeciesStructure):
    def __repr__(self):
        """
        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: a = S.structures(["a","b","c"]).random_element(); a
            {'a', 'b', 'c'}
        """
        s = GenericSpeciesStructure.__repr__(self)
        return "{"+s[1:-1]+"}"

    def canonical_label(self):
        """
        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: a = S.structures(["a","b","c"]).random_element(); a
            {'a', 'b', 'c'}
            sage: a.canonical_label()
            {'a', 'b', 'c'}
        """
        rng = list(range(1, len(self._labels) + 1))
        return SetSpeciesStructure(self.parent(), self._labels, rng)

    def transport(self, perm):
        """
        Return the transport of this set along the permutation perm.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: a = F.structures(["a", "b", "c"]).random_element(); a
            {'a', 'b', 'c'}
            sage: p = PermutationGroupElement((1,2))                                    # needs sage.groups
            sage: a.transport(p)                                                        # needs sage.groups
            {'a', 'b', 'c'}
        """
        return self

    def automorphism_group(self):
        """
        Return the group of permutations whose action on this set leave it
        fixed. For the species of sets, there is only one isomorphism
        class, so every permutation is in its automorphism group.

        EXAMPLES::

            sage: F = species.SetSpecies()
            sage: a = F.structures(["a", "b", "c"]).random_element(); a
            {'a', 'b', 'c'}
            sage: a.automorphism_group()                                                # needs sage.groups
            Symmetric group of order 3! as a permutation group
        """
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        return SymmetricGroup(max(1,len(self._labels)))


class SetSpecies(GenericCombinatorialSpecies, UniqueRepresentation):
    @staticmethod
    @accept_size
    def __classcall__(cls, *args, **kwds):
        """
        EXAMPLES::

            sage: E = species.SetSpecies(); E
            Set species
        """
        return super().__classcall__(cls, *args, **kwds)

    def __init__(self, min=None, max=None, weight=None):
        """
        Return the species of sets.

        EXAMPLES::

            sage: E = species.SetSpecies()
            sage: E.structures([1,2,3]).list()
            [{1, 2, 3}]
            sage: E.isotype_generating_series()[0:4]
            [1, 1, 1, 1]

            sage: S = species.SetSpecies()
            sage: c = S.generating_series()[0:3]
            sage: S._check()
            True
            sage: S == loads(dumps(S))
            True
        """
        GenericCombinatorialSpecies.__init__(self, min=min, max=max, weight=weight)
        self._name = "Set species"

    _default_structure_class = SetSpeciesStructure

    def _structures(self, structure_class, labels):
        """
        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: S.structures([1,2,3]).list()
            [{1, 2, 3}]
        """
        n = len(labels)
        yield structure_class(self, labels, range(1,n+1))

    _isotypes = _structures

    def _gs_callable(self, base_ring, n):
        r"""
        The generating series for the species of sets is given by
        `e^x`.

        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: g = S.generating_series()
            sage: [g.coefficient(i) for i in range(10)]
            [1, 1, 1/2, 1/6, 1/24, 1/120, 1/720, 1/5040, 1/40320, 1/362880]
            sage: [g.count(i) for i in range(10)]
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        """
        return base_ring(self._weight / factorial(n))

    def _itgs_list(self, base_ring, n):
        r"""
        The isomorphism type generating series for the species of sets is
        `\frac{1}{1-x}`.

        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: g = S.isotype_generating_series()
            sage: g[0:10]
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
            sage: [g.count(i) for i in range(10)]
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        """
        return base_ring(self._weight)

    def _cis(self, series_ring, base_ring):
        r"""
        The cycle index series for the species of sets is given by
        `exp\( \sum_{n=1}{\infty} = \frac{x_n}{n} \)`.

        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: g = S.cycle_index_series()                                            # needs sage.modules
            sage: g[0:5]                                                                # needs sage.modules
            [p[],
             p[1],
             1/2*p[1, 1] + 1/2*p[2],
             1/6*p[1, 1, 1] + 1/2*p[2, 1] + 1/3*p[3],
             1/24*p[1, 1, 1, 1] + 1/4*p[2, 1, 1] + 1/8*p[2, 2] + 1/3*p[3, 1] + 1/4*p[4]]
        """
        from .generating_series import ExponentialCycleIndexSeries
        res = ExponentialCycleIndexSeries(base_ring)

        if self.is_weighted():
            res *= self._weight

        return res


#Backward compatibility
SetSpecies_class = SetSpecies
